"""Metallurgy

A tool for manipulating alloy compositions, and calculation of their approximate
characteristics based on the linear mixture of elemental properties.
"""

from typing import Union, Optional, List

import pandas as pd
from elementy import PeriodicTable

from .alloy import Alloy
from .prototype import Prototype
from .calculate import (
    calculate,
    linear_mixture,
    deviation,
    range,
    maximum,
    minimum,
    get_property_function,
    get_all_properties,
)
from .plots import plot
from . import prototypes
from . import properties
from . import density
from . import valence
from . import entropy
from . import enthalpy
from . import radii
from . import viscosity
from . import price
from . import generate
from . import ratios
from . import plots
from . import analyse
from . import constants

# from . import rdf

periodic_table = PeriodicTable()
model = None


def set_model(model_in):
    """
    Sets a model to be used by metallurgy to provide predictions of alloy
    properties. Optional.

    Parameters
    ----------
    model_in : Cerebral model
        A model which takes alloy compositions as input, and produces
        material property predicitions as output.
    """
    global model
    model = model_in


def get_model():
    """Returns the model set for use in predicting material properties."""
    return model


def to_dataframe(
    alloys: Union[list, pd.DataFrame],
    data: Optional[dict] = None,
    properties: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Convert a list of alloys to a Pandas DataFrame

    :group: alloy.utils

    Parameters
    ----------
    alloys
        List of alloys to create a DataFrame
    data
        Dictionary with keys being property names and values being lists of
        property values
    properties
        List of properties to calculate for the alloys.

    """

    if data is not None and properties is not None:
        raise ValueError("Cannot set both data and properties arguments.")

    if not isinstance(alloys, pd.DataFrame):
        alloys_dataframe = pd.DataFrame.from_records(
            [alloy.to_dict() for alloy in alloys]
        )
    else:
        alloys_dataframe = alloys.copy()

    if data is not None:
        alloys_dataframe = pd.concat(
            [alloys_dataframe, pd.DataFrame.from_dict(data)],
            axis=1,
        )
    elif properties is not None:
        property_values = {}
        for p in properties:
            property_values[p] = calculate(alloys, p)

        alloys_dataframe = pd.concat(
            [alloys_dataframe, pd.DataFrame.from_dict(property_values)],
            axis=1,
        )

    return alloys_dataframe


def write_csv(
    alloys: Union[list, pd.DataFrame],
    data: Optional[dict] = None,
    properties: Optional[List[str]] = None,
    filename: str = "alloys",
):
    """Creates a CSV file of alloy data.

    :group: alloy.utils

    Parameters
    ----------
    alloys
        Dataset of alloys to create a CSV file.
    data
        Dictionary with keys being property names and values being lists of
        property values
    properties
        List of properties to calculate for the alloys.
    filename
        Name of the CSV file (will be suffixed with .csv).

    """

    if data is not None and properties is not None:
        raise ValueError("Cannot set both data and properties arguments.")

    if data is not None:
        to_dataframe(alloys, data=data).to_csv(filename + ".csv", index=False)

    elif properties is not None:
        to_dataframe(alloys, properties=properties).to_csv(
            filename + ".csv", index=False
        )


__all__ = [
    "periodic_table",
    "Alloy",
    "Prototype",
    "prototypes",
    "linear_mixture",
    "deviation",
    "range",
    "maximum",
    "minimum",
    "density",
    "valence",
    "entropy",
    "enthalpy",
    "radii",
    "price",
    "viscosity",
    "ratios",
    # "rdf",
    "plot",
    "plots",
    "generate",
    "calculate",
    "properties",
    "analyse",
    "constants",
    "model",
    "set_model",
    "get_model",
    "get_property_function",
    "get_all_properties",
    "write_csv",
]



"""Module enabling analysis of alloy data"""

import re
from typing import Tuple, Union, List
from numbers import Number

import numpy as np

import metallurgy as mg


def find_max(
    elements: List[str], property_name: str
) -> Union[Tuple[mg.Alloy, Number], None]:
    """Finds the maximum value of a property in an alloy system.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the alloy system.
    property_name
        The property to find the maximum of.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    system = mg.generate.system(elements, property_name=property_name)
    if system is not None:

        alloys, percentages, values = system

        max_index = np.argmax(values)

        return alloys[max_index], values[max_index]


def find_min(
    elements: List[str], property_name: str
) -> Union[Tuple[mg.Alloy, Number], None]:
    """Finds the minimum value of a property in an alloy system.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the alloy system.
    property_name
        The property to find the minimum of.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    system = mg.generate.system(elements, property_name=property_name)
    if system is not None:

        alloys, percentages, values = system

        min_index = np.argmin(values)

        return alloys[min_index], values[min_index]


def find_unique_elements(alloys: List[mg.Alloy]) -> List[str]:
    """Finds the unique elements in a list of alloy compositions.

    :group: alloy.utils

    Parameters
    ----------

    alloys
        List of alloys in which to find the unique elements.

    """

    unique_elements = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in unique_elements:
                unique_elements.append(element)

    return unique_elements


def find_unique_percentages(alloys: List[mg.Alloy]) -> dict:
    """Finds the unique percentages per element in a list of alloy compositions.

    :group: alloy.utils

    Parameters
    ----------

    alloys
        List of alloys in which to find the unique percentages.

    """

    unique_elements = find_unique_elements(alloys)
    percentages = {e: [] for e in unique_elements}
    for alloy in alloys:
        for element in alloy.composition:
            percentage = alloy.composition[element]
            if percentage not in percentages[element]:
                percentages[element].append(percentage)

    return percentages


def composition_weighted_average(alloys, data):
    unique_elements = find_unique_elements(alloys)

    per_element_property = {e: 0 for e in unique_elements}
    per_element_total = {e: 0 for e in unique_elements}
    for i, alloy in enumerate(alloys):
        for element in alloy.composition:
            if data[i] is None:
                continue

            per_element_property[element] += (
                data[i] * alloy.composition[element]
            )
            per_element_total[element] += alloy.composition[element]

    for element in per_element_property:
        if per_element_total[element] > 0:
            per_element_property[element] /= per_element_total[element]
        else:
            per_element_property[element] = None

    return per_element_property



"""Module providing enthalpy related calculations."""

from collections.abc import Iterable
from numbers import Number
from typing import Tuple, Union, List

import numpy as np

import metallurgy as mg


def gamma(element_a: str, element_b: str) -> Union[Number, None]:
    """Calculates the gamma term of the Miedema model.
    See equation 1 of: http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A
    element_b
        The periodic table symbol of element B
    """

    Q, P, R = calculate_QPR(element_a, element_b)
    if R is not None:
        return (
            calculate_electronegativity_enthalpy_component(
                element_a, element_b, P
            )
            + calculate_WS_enthalpy_component(element_a, element_b, Q)
            - R
        )

    return None


def calculate_QPR(
    element_a: str, element_b: str
) -> Union[Tuple[Number, Number, Number], None]:
    """Calculates the Q, P, and R factors of the Miedema model.
    See equation 1 of: http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A
    element_b
        The periodic table symbol of element B
    """

    series_a = mg.periodic_table.elements[element_a]["series"]
    if element_a in ("Ca", "Sr", "Ba"):
        series_a = "non_transition_metal"

    series_b = mg.periodic_table.elements[element_b]["series"]
    if element_b in ("Ca", "Sr", "Ba"):
        series_b = "non_transition_metal"

    if series_a == "transition_metal" and series_b == "transition_metal":
        P = 14.1
        R = 0
    elif series_a != "transition_metal" and series_b != "transition_metal":
        P = 10.6
        R = 0
    else:
        P = 12.3

        if (
            mg.periodic_table.elements[element_a]["miedema_R"] is not None
            and mg.periodic_table.elements[element_b]["miedema_R"] is not None
        ):
            R = (
                mg.periodic_table.elements[element_a]["miedema_R"]
                * mg.periodic_table.elements[element_b]["miedema_R"]
            )
        else:
            R = None

    Q = P * 9.4

    return Q, P, R


def calculate_electronegativity_enthalpy_component(
    element_a: str, element_b: str, P: Number
) -> Number:
    """Calculates the electronegativity contribution to the gamma factor in the
    Miedema model of mixing enthalpy.  See equation 1 of:
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A
    element_b
        The periodic table symbol of element B
    P
        An empirical factor dependent on the kinds of elements being
        mixed. P=14.1 for two transition metals, P=10.7 for two non-transition
        metals, and P=12.35 for one of each kind.
    """

    electronegativity_difference = (
        mg.periodic_table.elements[element_a]["electronegativity_miedema"]
        - mg.periodic_table.elements[element_b]["electronegativity_miedema"]
    )
    return -P * (electronegativity_difference**2)


def calculate_WS_enthalpy_component(
    element_a: str, element_b: str, Q: Number
) -> Number:
    """Calculates the Wigner-Seitz radius discontinuity contribution to the
    gamma factor in the Miedema model of mixing enthalpy.  See equation 1 of:
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A
    element_b
        The periodic table symbol of element B
    Q
        An empirical factor dependent on the kinds of elements being
        mixed. Dependent on the P factor discussed, Q = 9.4*P, discussed in
        documentation of
        :func:`~metallurgy.enthalpy.calculate_electronegativity_enthalpy_component`.
    """

    return Q * (
        wigner_seitz_electron_density_discontinuity_delta(element_a, element_b)
        ** 2
    )


def wigner_seitz_electron_density_discontinuity_delta(
    element_a: str, element_b: str
) -> Number:
    """Calculates the Wigner-Seitz radius discontinuity contribution to the
    gamma factor in the Miedema model of mixing enthalpy.  See equation 1 of:
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A
    element_b
        The periodic table symbol of element B
    """

    density_a = mg.periodic_table.elements[element_a][
        "wigner_seitz_electron_density"
    ]
    density_b = mg.periodic_table.elements[element_b][
        "wigner_seitz_electron_density"
    ]
    return (density_a ** (1.0 / 3.0)) - (density_b ** (1.0 / 3.0))


def calculate_surface_concentration(
    elements: List[str], volumes: List[Number], composition: dict
) -> Number:
    """Calculates the surface concentration in the Miedema model of mixing
    enthalpy.  See equation 2 of: http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    elements
        The elements being used to calculate the surface concentration.
    volumes
        The atomic volumes of the elements.
    composition
        Dictionary containing the percentage of each element in the surface
        composition.
    """

    reduced_vol_a = composition[elements[0]] * (volumes[0] ** (2.0 / 3.0))
    reduced_vol_b = composition[elements[1]] * (volumes[1] ** (2.0 / 3.0))

    return reduced_vol_a / (reduced_vol_a + reduced_vol_b)


def calculate_corrected_volume(
    element_a: str, element_b: str, surface_concentration_a: Number
):
    """Calculates the corrected volume of element_a in an alloy with element_b,
    as defined by the Meidema model.  See equation 2 of:
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The element for which to correct the volume.
    element_b
        The other element in the alloy.
    surface_concentration_a
        The surface concentration of element_a, as calculated by:
        :func:`~metallurgy.enthalpy.calculate_surface_concentration`.

    """

    pure_volume_a = mg.periodic_table.elements[element_a]["volume_miedema"]

    electronegativity_difference = (
        mg.periodic_table.elements[element_a]["electronegativity_miedema"]
        - mg.periodic_table.elements[element_b]["electronegativity_miedema"]
    )

    valence_factor = None
    if element_a in ["Ca", "Sr", "Ba"]:
        valence_factor = 0.04
    elif element_a in ["Ru", "Rh", "Pd", "Os", "Ir", "Pt", "Au"]:
        valence_factor = 0.07

    if valence_factor is None:
        if mg.periodic_table.elements[element_a]["series"] == "alkaliMetal":
            valence_factor = 0.14
        elif mg.periodic_table.elements[element_a]["valence_electrons"] == 2:
            valence_factor = 0.1
        elif mg.periodic_table.elements[element_a]["valence_electrons"] == 3:
            valence_factor = 0.07
        else:
            valence_factor = 0.04

    f_AB = 1 - surface_concentration_a

    corrected_volume_a = (pure_volume_a ** (2.0 / 3.0)) * (
        1 + valence_factor * f_AB * electronegativity_difference
    )

    return corrected_volume_a


def calculate_interface_enthalpy(
    element_a: str, element_b: str, volume_a: Number
) -> Number:
    """Calculates the interfacial enthalpy term of the Miedema model.  See
    equation 1 of: http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    element_a
        The periodic table symbol of element A.
    element_b
        The periodic table symbol of element B.
    volume_a
        The atomic volume of element A.

    """

    density_a = mg.periodic_table.elements[element_a][
        "wigner_seitz_electron_density"
    ]
    density_b = mg.periodic_table.elements[element_b][
        "wigner_seitz_electron_density"
    ]

    _gamma = gamma(element_a, element_b)
    if _gamma is not None:
        interface_enthalpy = (
            2
            * volume_a
            * _gamma
            / (density_a ** (-1.0 / 3.0) + density_b ** (-1.0 / 3.0))
        )

        return interface_enthalpy


def mixing_enthalpy(alloy: Union[mg.Alloy, str, dict]):
    """Calculates the Miedema model mixing enthalpy.  See equation 15a of:
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the mixing enthalpy of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_enthalpy(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    if alloy.num_elements > 1:

        element_pairs = [
            (a, b)
            for idx, a in enumerate(alloy.elements)
            for b in alloy.elements[idx + 1 :]
        ]

        total_mixing_enthalpy = 0
        for pair in element_pairs:
            tmp_composition = {}
            sub_composition = 0
            for element in pair:
                sub_composition += alloy.composition[element]
            for element in pair:
                tmp_composition[element] = (
                    alloy.composition[element] / sub_composition
                )

            surface_concentration_a = None
            volume_a = mg.periodic_table.elements[pair[0]]["volume_miedema"]
            volume_b = mg.periodic_table.elements[pair[1]]["volume_miedema"]

            if volume_a is None or volume_b is None:
                return None

            for _ in range(10):

                surface_concentration_a = calculate_surface_concentration(
                    pair, [volume_a, volume_b], tmp_composition
                )

                volume_a = calculate_corrected_volume(
                    pair[0], pair[1], surface_concentration_a
                )

                volume_b = calculate_corrected_volume(
                    pair[1], pair[0], 1 - surface_concentration_a
                )

            interface_ab = calculate_interface_enthalpy(
                pair[0], pair[1], volume_a
            )
            interface_ba = calculate_interface_enthalpy(
                pair[1], pair[0], volume_b
            )

            if interface_ab is not None and interface_ba is not None:

                chemical_enthalpy = (
                    alloy.composition[pair[0]]
                    * alloy.composition[pair[1]]
                    * (
                        (1 - surface_concentration_a) * interface_ab
                        + surface_concentration_a * interface_ba
                    )
                )
                total_mixing_enthalpy += chemical_enthalpy

            else:
                return None

    else:
        total_mixing_enthalpy = 0.0

    return total_mixing_enthalpy


def mixing_Gibbs_free_energy(alloy: Union[mg.Alloy, str, dict]) -> Number:
    """Calculates the Gibbs free energy of mixing.

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the Gibbs free energy of mixing of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_Gibbs_free_energy(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mix_enthalpy = mixing_enthalpy(alloy)
    melting_temperature = mg.linear_mixture(alloy, "melting_temperature")
    mixing_entropy = mg.entropy.mixing_entropy(alloy)

    if (
        mix_enthalpy is None
        or melting_temperature is None
        or mixing_entropy is None
    ):
        return None

    return (
        (mix_enthalpy * 1e3)
        - melting_temperature * mixing_entropy * mg.constants.idealGasConstant
    )


def topological_enthalpy(alloy):
    """Calculates the topological enthalpy. See equation 16b of
    http://dx.doi.org/10.1016/j.cpc.2016.08.013

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the topological enthalpy of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mismatch_PHS(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    te = 0
    for element in alloy.composition:
        te += (
            mg.periodic_table.elements[element]["fusion_enthalpy"]
            * alloy.composition[element]
        )

    return te


def mismatch_PHS(alloy: Union[mg.Alloy, str, dict]) -> Number:
    """Calculates the mismatch PHS factor. See
    https://doi.org/10.1016/j.intermet.2012.11.020.

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the PHS factor of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mismatch_PHS(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mix_enthalpy = mixing_enthalpy(alloy)
    mismatch_entropy = mg.entropy.mismatch_entropy(alloy)

    if mix_enthalpy is not None and mismatch_entropy is not None:
        return mix_enthalpy * mismatch_entropy

    return None


def mixing_PHS(alloy):
    """Calculates the mixing PHS factor. See
    https://doi.org/10.1016/j.intermet.2012.11.020.

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the PHS factor of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_PHS(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mix_enthalpy = mixing_enthalpy(alloy)
    mixing_entropy = mg.entropy.mixing_entropy(alloy)

    if mix_enthalpy is not None and mixing_entropy is not None:
        return mix_enthalpy * mixing_entropy

    return None


def mixing_PHSS(alloy: Union[mg.Alloy, str, dict]) -> Number:
    """Calculates the mismatch PHS factor. See
    https://doi.org/10.1016/j.intermet.2012.11.020.

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the PHSS factor of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_PHSS(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mix_enthalpy = mixing_enthalpy(alloy)
    mixing_entropy = mg.entropy.mixing_entropy(alloy)
    mismatch_entropy = mg.entropy.mismatch_entropy(alloy)

    if (
        mix_enthalpy is None
        or mixing_entropy is None
        or mismatch_entropy is None
    ):
        return None

    return mix_enthalpy * mixing_entropy * mismatch_entropy


def thermodynamic_factor(alloy: Union[mg.Alloy, str, dict]) -> Number:
    """Calculates the thermodynamic factor. See equation 11 of
    https://doi.org/10.1016/j.matdes.2020.108835.

    :group: calculations.enthalpy

    Parameters
    ----------

    alloy
        Alloy to calculate the thermodynamic factor of mixing of.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [thermodynamic_factor(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    melting_temperature = mg.linear_mixture(alloy, "melting_temperature")
    mix_enthalpy = mixing_enthalpy(alloy)
    mixing_entropy = mg.entropy.mixing_entropy(alloy)

    if (
        melting_temperature is None
        or mix_enthalpy is None
        or mixing_entropy is None
    ):
        return None

    return (melting_temperature * mixing_entropy) / (
        np.abs(mix_enthalpy * 1e3) + 1e-10
    )



"""``metallurgy.constants``
=============================

Module providing some physical constants used in calculation functions.

"""


plankConstant = 6.63e-34
avogadroNumber = 6.022e23
idealGasConstant = 8.31
boltzmann = 1.38064852e-23



"""Module providing density related calculations."""

from typing import Union, List
from collections.abc import Iterable
from numbers import Number

import metallurgy as mg


def theoretical_density(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the theoretical density of an alloy.  See equation 8 of
    https://doi.org/10.1016/j.jnoncrysol.2019.03.001.

    :group: calculations.density

    Parameters
    ----------

    alloy
        The alloy for which to calculate the theoretcial density.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [theoretical_density(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mass_fractions = {}
    masses = {}
    total_mass = 0
    for element in alloy.elements:
        masses[element] = mg.periodic_table.elements[element]["mass"]
        total_mass += masses[element] * alloy.composition[element]

    for element in alloy.elements:
        mass_fractions[element] = (
            alloy.composition[element] * masses[element] / total_mass
        )

    total = 0
    for element in alloy.elements:
        total += (
            mass_fractions[element]
            / mg.periodic_table.elements[element]["density"]
        )

    return 1 / total



"""Module enabling calculation of radius related alloy properties."""

from collections.abc import Iterable
from typing import List, Union

import numpy as np

import metallurgy as mg


def radius_gamma(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the radius gamma property for an alloy.

    :group: calculations.radii

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the linear mixture.

    """

    # If a list of alloys is given, return a list of data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [radius_gamma(a) for a in alloy]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    max_radius = 0
    min_radius = 1000
    radii = {}
    for element in alloy.elements:
        radii[element] = mg.periodic_table.elements[element]["radius"]
        if radii[element] is None:
            return None

        if radii[element] > max_radius:
            max_radius = radii[element]
        if radii[element] < min_radius:
            min_radius = radii[element]

    mean_radius = 0
    for element in alloy.elements:
        mean_radius += alloy.composition[element] * radii[element]

    r_min_delta_sq = (min_radius + mean_radius) ** 2
    r_max_delta_sq = (max_radius + mean_radius) ** 2
    r_mean_sq = mean_radius**2

    numerator = 1.0 - np.sqrt((r_min_delta_sq - r_mean_sq) / (r_min_delta_sq))
    denominator = 1.0 - np.sqrt(
        (r_max_delta_sq - r_mean_sq) / (r_max_delta_sq)
    )

    return numerator / denominator


def lattice_distortion(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the lattice distortion of an alloy.

    :group: calculations.radii

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the linear mixture.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [lattice_distortion(a) for a in alloy]
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    mean_radius = 0
    for element in alloy.elements:
        radius = mg.periodic_table.elements[element]["radius"]
        if radius is None:
            return None

        mean_radius += alloy.composition[element] * radius

    _lattice_distortion = 0
    for i in range(len(alloy.elements) - 1):
        for j in range(i + 1, len(alloy.elements)):
            element = alloy.elements[i]
            other_element = alloy.elements[j]

            radius_a = (
                alloy.composition[element]
                * mg.periodic_table.elements[element]["radius"]
            )
            radius_b = (
                alloy.composition[element]
                * mg.periodic_table.elements[other_element]["radius"]
            )

            _lattice_distortion += (
                alloy.composition[element]
                * alloy.composition[other_element]
                * np.abs(radius_a + radius_b - 2 * mean_radius)
            ) / (2 * mean_radius)

    return _lattice_distortion



import metallurgy as mg


def test_calculate_property_name():
    assert mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, "melting_temperature_linearmix"
    ) == mg.linear_mixture({"Cu": 0.5, "Zr": 0.5}, "melting_temperature")

    assert mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, "melting_temperature_deviation"
    ) == mg.deviation({"Cu": 0.5, "Zr": 0.5}, "melting_temperature")


def test_calculate_property_name_list():

    multi_calc = mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, ["melting_temperature", "mass"]
    )

    assert len(multi_calc) == 2
    assert "melting_temperature" in multi_calc and "mass" in multi_calc

    assert mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, ["melting_temperature_linearmix"]
    )["melting_temperature_linearmix"] == mg.linear_mixture(
        {"Cu": 0.5, "Zr": 0.5}, "melting_temperature"
    )



"""``metallurgy.price``
=============================

Module enabling calculation of approximate alloy price.

"""

from typing import Union, List
from collections.abc import Iterable

import metallurgy as mg


def price(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the approximate price per kilogramme of an alloy.

    :group: calculations.price

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the linear mixture.

    """

    # If a list of alloys is given, return a list of price data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [price(a) for a in alloy]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    # Get the total mass to enable conversion of atomic percent to weight percent
    total_mass = mg.linear_mixture(alloy, "mass")

    total_price = 0
    for element in alloy.elements:
        # Only attempt calculation if pricing data is available
        if mg.periodic_table.elements[element]["price"] is not None:
            # Convert atomic percent to weight percent
            weight_percent = (
                mg.periodic_table.elements[element]["mass"]
                * alloy.composition[element]
                / total_mass
            )

            # Add weighted contribution to the price approximation
            total_price += (
                weight_percent * mg.periodic_table.elements[element]["price"]
            )
        else:
            return None

    return total_price



import metallurgy as mg
import pytest


def test_random_alloy(random_alloy=None):
    if random_alloy is None:
        random_alloy = mg.generate.random_alloy()
    assert sum(random_alloy.composition.values()) == pytest.approx(1.0)


def test_random_alloys():

    num_alloys = 100
    random_alloys = mg.generate.random_alloys(num_alloys)

    assert len(random_alloys) == num_alloys

    for random_alloy in random_alloys:
        test_random_alloy(random_alloy)


def test_generate_binary_alloys():

    elements = ["Cu", "Zr"]

    step = 1
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 101
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}

    step = 2
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 51
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}

    step = 0.5
    alloys, percentages = mg.generate.binary(elements, step)
    assert len(alloys) == len(percentages) == 201
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[1]: 1.0}


def test_generate_ternary_alloys():

    elements = ["Cu", "Zr", "Al"]

    step = 1
    alloys, percentages = mg.generate.ternary(elements, step)
    assert len(alloys) == len(percentages) == 5151
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[2]: 1.0}

    step = 2
    alloys, percentages = mg.generate.ternary(elements, step)
    assert len(alloys) == len(percentages) == 1326
    assert alloys[0].composition == {elements[0]: 1.0}
    assert alloys[-1].composition == {elements[2]: 1.0}


def check_constraints(alloys, constraints):
    for alloy in alloys:
        for element in constraints["percentages"]:
            if constraints["percentages"][element]["min"] > 0:
                assert element in alloy.composition

            if element in alloy.composition:
                assert (
                    alloy.composition[element]
                    >= constraints["percentages"][element]["min"]
                )
                assert (
                    alloy.composition[element]
                    <= constraints["percentages"][element]["max"]
                )

                if "precedence" in constraints["percentages"][element]:
                    for other_element in constraints["percentages"]:
                        if element == other_element:
                            continue

                        if other_element in alloy.composition:
                            if (
                                "precedence"
                                in constraints["percentages"][other_element]
                            ):
                                if (
                                    constraints["percentages"][element][
                                        "precedence"
                                    ]
                                    > constraints["percentages"][
                                        other_element
                                    ]["precedence"]
                                ):
                                    assert (
                                        alloy.composition[element]
                                        >= alloy.composition[other_element]
                                    )

            else:
                assert constraints["percentages"][element]["min"] == 0


def test_constraints():

    num_alloys = 100

    constraint_sets = [
        {
            "percentages": {
                "Cu": {"min": 0.2, "max": 0.6},
                "Fe": {"min": 0.0, "max": 1.0},
                "Ni": {"min": 0.01, "max": 0.99, "precedence": 1},
            },
            "max_elements": 10,
            "min_elements": 1,
            "percentage_step": 0.01,
        },
        {
            "percentages": {
                "Zr": {"min": 0.45, "max": 0.65},
                "Ti": {"min": 0.0, "max": 0.2},
                "Cu": {"min": 0.0, "max": 0.3},
                "Ni": {"min": 0.0, "max": 0.3},
                "Al": {"min": 0.0, "max": 0.2},
            },
            "max_elements": 5,
            "min_elements": 1,
            "percentage_step": 0.005,
        },
        {
            "percentages": {
                "Zr": {"min": 0.45, "max": 0.65},
                "Ti": {"min": 0.0, "max": 0.8},
            },
            "max_elements": 2,
            "min_elements": 2,
            "percentage_step": 0.005,
        },
    ]

    for constraint in constraint_sets:
        random_alloys = mg.generate.random_alloys(
            num_alloys,
            min_elements=constraint["min_elements"],
            max_elements=constraint["max_elements"],
            percentage_constraints=constraint["percentages"],
            percentage_step=constraint["percentage_step"],
        )

        check_constraints(random_alloys, constraint)


def test_mixture():

    A = mg.Alloy("Cu50Zr50")

    mixed = mg.generate.mixture([A])

    assert mixed.composition == A.composition

    B = mg.Alloy("Fe50Ni50")

    mixed = mg.generate.mixture([A, B])

    assert mixed.composition == {
        "Cu": 0.25,
        "Zr": 0.25,
        "Fe": 0.25,
        "Ni": 0.25,
    }

    mixed = mg.generate.mixture([A, B], [0.9, 0.1])

    assert mixed.composition == {
        "Cu": 0.45,
        "Zr": 0.45,
        "Fe": 0.05,
        "Ni": 0.05,
    }

    C = mg.Alloy(
        "Cu50Zr50",
        constraints={
            "percentages": {"Cu": {"min": 0.1, "max": 0.6}},
            "max_elements": 2,
            "min_elements": 1,
        },
    )

    mixed = mg.generate.mixture([C])

    assert mixed.composition == C.composition

    mixed = mg.generate.mixture([C, B])

    assert len(mixed.elements) <= 2
    assert mixed.composition["Cu"] <= 0.6
    assert mixed.composition["Cu"] >= 0.1

    assert mg.generate.mixture(["Cu[A1]", "Fe[A2]"]) == "Cu50Fe50[A1]"
    assert mg.generate.mixture(["Cu[A2]", "Fe[A1]"]) == "Cu50Fe50[A2]"



from typing import List, Optional

import matplotlib as mpl
import matplotlib.collections as mcoll
import matplotlib.pyplot as plt
import numpy as np
import ternary as ternary_plt
from bokeh.io import export_png
from bokeh.io import show as show_
from bokeh.models import (BasicTicker, ColorBar, ColumnDataSource,
                          LinearColorMapper, LogColorMapper)
from bokeh.plotting import figure
from bokeh.sampledata.periodic_table import elements
from bokeh.transform import dodge
from matplotlib.figure import Figure

import metallurgy as mg


def plot(
    alloys: Optional = None,
    data: Optional[list] = [],
    elements: Optional[List[str]] = None,
    property_name: str = None,
    label: str = None,
    labels: Optional[List[str]] = None,
    save_path: str = None,
    step: Optional = 1,
):
    """An automatic interface to metallurgy's alloy system plotting
    functions.

    Parameters
    ----------

    alloys
        List of alloys being plotted.
    data
        List of data belonging to alloys to plot.
    label
        String for labelling the plotted data.

    """

    if (alloys is None and elements is None) or (
        alloys is not None and elements is not None
    ):
        raise ValueError("Must provide either alloys or elements.")

    if elements is not None:
        if len(elements) < 4:
            alloys, percentages = mg.generate.system(elements, step=step)
        else:
            alloys = mg.generate.system(elements, step=step)

    if len(data) == 0 and property_name is None:
        raise ValueError("Must provide either data or a property_name.")
    elif property_name is not None:
        data = mg.calculate(alloys, property_name)

    if isinstance(alloys[0], mg.Alloy):
        num_elements = len(mg.analyse.find_unique_elements(alloys))
    elif isinstance(alloys[0], list) and isinstance(alloys[0][0], mg.Alloy):
        num_elements = len(
            mg.analyse.find_unique_elements(
                [alloy for sublist in alloys for alloy in sublist]
            )
        )
    else:
        raise ValueError("Could not determine number of elements.")

    if step is not None:
        step /= 100

    if num_elements == 2:
        return binary(
            alloys, data, ylabel=label, labels=labels, save_path=save_path
        )
    elif num_elements == 3:
        return ternary(
            alloys, data, label=label, save_path=save_path, step=step
        )
    elif num_elements == 4:
        return quaternary(
            alloys, data, label=label, save_path=save_path, step=step
        )
    else:
        raise NotImplementedError(
            "No plotting available for "
            + str(num_elements)
            + " element alloys."
        )


def binary(
    alloys,
    data,
    xlabel: str = None,
    ylabel: str = None,
    labels: List[str] = None,
    scatter_data: list = None,
    use_colorline: bool = False,
    save_path: str = None,
):
    """Plots an alloy property across a binary alloy composition.

    :group: plots

    Parameters
    ----------

    alloys : List[mg.Alloy, str, dict]
        The alloys across the binary composition
    """

    fig = Figure()
    ax1 = fig.add_subplot(111)

    elements = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in elements:
                elements.append(element)

    percentages = []
    for alloy in alloys:
        if elements[0] in alloy.composition:
            percentages.append(alloy.composition[elements[0]] * 100)
        else:
            percentages.append(0)

    percentages, data = zip(*sorted(zip(percentages, data)))

    if use_colorline:
        lc = colorline(data[0], data[1], ax1, z=percentages)
        cbar = plt.colorbar(lc)
        cbar.set_label(elements[0] + " %", rotation=270)

    elif isinstance(data[0], (list, np.ndarray)):
        for i in range(len(data[0])):
            label = None
            if labels is not None:
                label = labels[i]

            ax1.plot(percentages, [d[i] for d in data], label=label)

    elif isinstance(data[0], tuple):
        if isinstance(data[0][0], list):
            for i in range(len(data[0][0])):
                label = None
                if labels is not None:
                    label = labels[i]

                ax1.plot(percentages, [d[0][i] for d in data], label=label)
                ax1.fill_between(
                    percentages,
                    [d[0][i] - d[1][i] for d in data],
                    [d[0][i] + d[1][i] for d in data],
                    alpha=0.5,
                )

        else:
            ax1.plot(percentages, [d[0] for d in data])
            ax1.fill_between(
                percentages,
                [d[0] - d[1] for d in data],
                [d[0] + d[1] for d in data],
                alpha=0.5,
            )

    else:
        ax1.plot(percentages, data)

    if scatter_data is not None:
        for scatter_datum in scatter_data:
            ax1.scatter(
                scatter_datum["data"][0],
                scatter_datum["data"][1],
                marker=scatter_datum["marker"],
                label=scatter_datum["label"],
                edgecolors="k",
                zorder=20,
            )
        ax1.legend(loc="best")
    elif labels is not None:
        ax1.legend(loc="best")

    ax1.autoscale()

    if not isinstance(data[0], list) and not use_colorline:
        ax2 = ax1.twiny()
        ax2.set_xticks(ax1.get_xticks())
        ax2.set_xbound(ax1.get_xbound())
        ax2.set_xticklabels([int(100 - x) for x in ax1.get_xticks()])

        ax1.set_xlabel(elements[0] + " %")
        ax2.set_xlabel(elements[1] + " %")
        ax2.grid(False)

    if xlabel is not None:
        ax1.set_xlabel(xlabel)
    if ylabel is not None:
        ax1.set_ylabel(ylabel)

    fig.tight_layout()

    # fig.gcf().set_dpi(300)
    if save_path is not None:
        fig.savefig(save_path)
    else:
        fig.show()

    fig.clf()


def ternary(
    alloys,
    data,
    step=0.01,
    scatter_data=None,
    save_path=None,
    label=None,
    title=None,
    vmin=None,
    vmax=None,
    ax=None,
    showColorbar=True,
):
    """Plots an alloy property across a ternary alloy composition.

    :group: plots

    Parameters
    ----------

    alloys : List[mg.Alloy, str, dict]
        The alloys across the ternary composition
    """

    scale = 1 / step
    multiple = 0.1 / step
    fontsize = 10
    tick_fontsize = 6
    tick_offset = 0.018
    gridline_width = 0.3
    gridline_style = "--"
    gridline_color = "white"

    elements = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in elements:
                elements.append(element)

    if not isinstance(data, dict):
        data = list(data)

        heatmap_data = dict()

        for i in range(len(data)):
            percent_A = 0
            percent_B = 0

            if elements[0] in alloys[i].composition:
                percent_A = round(alloys[i].composition[elements[0]] / step, 2)
            if elements[1] in alloys[i].composition:
                percent_B = round(alloys[i].composition[elements[1]] / step, 2)

            heatmap_data[(percent_A, percent_B)] = data[i]

        data = heatmap_data

    if ax is None:
        figure, tax = ternary_plt.figure(scale=scale)
    else:
        figure, tax = ternary_plt.figure(scale=scale, ax=ax)

    tax.get_axes().axis("off")

    tax.gridlines(
        color=gridline_color,
        multiple=multiple,
        linewidth=gridline_width,
        ls=gridline_style,
    )

    tax.set_axis_limits({"b": [0, 100], "l": [0, 100], "r": [0, 100]})
    tax.get_ticks_from_axis_limits(multiple=multiple)
    tax.set_custom_ticks(
        fontsize=tick_fontsize, offset=tick_offset, multiple=multiple
    )

    tax.left_axis_label(elements[2] + " %", fontsize=fontsize, offset=0.12)
    tax.right_axis_label(elements[1] + " %", fontsize=fontsize, offset=0.12)
    tax.bottom_axis_label(elements[0] + " %", fontsize=fontsize, offset=0.12)
    tax.clear_matplotlib_ticks()

    tax.set_title(title, pad=15)

    viridis_cmap = mpl.cm.get_cmap("viridis")
    tax.heatmap(
        heatmap_data,
        cmap=viridis_cmap,
        vmax=vmax,
        vmin=vmin,
        cbarlabel=label,
        colorbar=showColorbar,
    )

    if scatter_data is not None:
        for scatter_datum in scatter_data:
            tax.scatter(
                scatter_datum["data"],
                marker=scatter_datum["marker"],
                label=scatter_datum["label"],
                edgecolors="k",
                zorder=20,
            )

        tax.legend(loc="upper right", handletextpad=0.1, frameon=False)

    tax.get_axes().set_aspect(1)
    tax._redraw_labels()
    # plt.gcf().set_dpi(300)

    if ax is None:
        if save_path is None:
            tax.show()
        else:
            tax.savefig(save_path)

        tax.close()
        figure.clf()


def quaternary(
    quaternary_alloys,
    data,
    label,
    save_path: Optional[str] = None,
    step: Optional[float] = 0.01,
):

    unique_percentages = mg.analyse.find_unique_percentages(
        quaternary_alloys[-1]
    )

    quaternary_element = list(unique_percentages.keys())[
        np.argmin([len(unique_percentages[e]) for e in unique_percentages])
    ]
    ternary_elements = []
    for element in unique_percentages:
        if element != quaternary_element:
            ternary_elements.append(element)

    vmin = min(y for x in data for y in x)
    vmax = max(y for x in data for y in x)

    columns = int(np.ceil(np.sqrt(len(quaternary_alloys))))
    rows = int(np.ceil(len(quaternary_alloys) / columns))
    numGridCells = columns * rows
    gridExcess = numGridCells - len(quaternary_alloys)

    fig = Figure(figsize=(4 * columns, 4 * rows))

    lastAx = None
    for i in reversed(range(len(quaternary_alloys))):
        iRow = i // columns
        iCol = i % columns

        ax = plt.subplot2grid((rows, columns), (iRow, iCol))

        if gridExcess != 0 and iRow == (rows - 1):
            if gridExcess % 2 == 0:
                ax = plt.subplot2grid((rows, columns), (iRow, iCol + 1))
            else:
                ax = plt.subplot2grid(
                    (rows, columns * 2), (iRow, 1 + (iCol * 2)), colspan=2
                )

        if lastAx is None:
            lastAx = ax

        unique_percentages = mg.analyse.find_unique_percentages(
            quaternary_alloys[i]
        )

        if quaternary_element in unique_percentages:
            quaternary_percentage = (
                unique_percentages[quaternary_element][0] * 100
            )
        else:
            quaternary_percentage = 0.0

        remaining_percentage = round(100 - quaternary_percentage, 2)
        remaining_percentage_str = pretty_percentage(str(remaining_percentage))
        quaternary_percentage_str = pretty_percentage(
            str(round(quaternary_percentage, 2))
        )

        title = (
            "("
            + "".join(ternary_elements)
            + ")$_{"
            + remaining_percentage_str
            + "}$"
            + quaternary_element
            + "$_{"
            + quaternary_percentage_str
            + "}$"
        )

        ternary_alloys = quaternary_alloys[i][:]
        for j in range(len(ternary_alloys)):
            ternary_alloys[j].remove_element(quaternary_element)

        ternary(
            ternary_alloys,
            data[i],
            ax=ax,
            vmin=vmin,
            vmax=vmax,
            title=title,
            label=label,
            showColorbar=False,
            step=step,
        )

    viridis_cmap = mpl.cm.get_cmap("viridis")

    cax = fig.add_axes(
        [
            lastAx.get_position().x1 + 0.01,
            lastAx.get_position().y0 + 0.03,
            0.0075,
            lastAx.get_position().height,
        ]
    )

    colorbar = fig.colorbar(
        mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=vmin, vmax=vmax),
            cmap=viridis_cmap,
        ),
        cax=cax,
    )
    colorbar.set_label(label, labelpad=20, rotation=270)

    plt.gcf().set_dpi(300)
    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)
    plt.clf()
    plt.cla()
    plt.close()


def colorline(
    x,
    y,
    ax,
    z=None,
    cmap="viridis",
    norm=plt.Normalize(0.0, 100.0),
    linewidth=2,
    alpha=1.0,
):
    """
    http://nbviewer.ipython.org/github/dpsanders/matplotlib-examples/blob/master/colorline.ipynb
    http://matplotlib.org/examples/pylab_examples/multicolored_line.html
    Plot a colored line with coordinates x and y
    Optionally specify colors in the array z
    Optionally specify a colormap, a norm function and a line width
    """

    # Default colors equally spaced on [0,1]:
    if z is None:
        z = np.linspace(0.0, 1.0, len(x))

    # Special case if a single number:
    # to check for numerical input -- this is a hack
    if not hasattr(z, "__iter__"):
        z = np.array([z])

    z = np.asarray(z)

    segments = make_segments(x, y)

    lc = mcoll.LineCollection(
        segments,
        array=z,
        cmap=cmap,
        norm=norm,
        linewidth=linewidth,
        alpha=alpha,
    )

    ax.add_collection(lc)
    return lc


def make_segments(x, y):
    """
    Create list of line segments from x and y coordinates, in the correct format
    for LineCollection: an array of the form numlines x (points per line) x 2 (x
    and y) array
    """

    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    return segments


def pretty_percentage(string: str) -> str:
    """Trim trailing decimals from percentage string"""
    while string[-1] == "0":
        string = string[:-1]
    if string[-1] == ".":
        string = string[:-1]
    return string


def periodic_table(
    data,
    show: bool = True,
    output_filename: Optional[str] = None,
    width: int = 1050,
    cmap: str = "viridis",
    alpha: float = 0.65,
    extended: bool = True,
    periods_remove: List[int] = None,
    groups_remove: List[int] = None,
    log_scale: bool = False,
    cbar_height: float = None,
    cbar_standoff: int = 12,
    cbar_fontsize: int = 14,
    blank_color: str = "#c4c4c4",
    under_value: float = None,
    under_color: str = "#140F0E",
    over_value: float = None,
    over_color: str = "#140F0E",
    special_elements: List[str] = None,
    special_color: str = "#6F3023",
):
    """
    Plot a heatmap over the periodic table of elements.
    Adapted from work of Andrew Rosen at
    https://github.com/arosen93/ptable_trends

    Parameters
    ----------
    filename : str
        Path to the .csv file containing the data to be plotted.
    show : str
        If True, the plot will be shown.
    output_filename : str
        If not None, the plot will be saved to the specified (.html) file.
    width : float
        Width of the plot.
    cmap : str
        plasma, inferno, viridis, magma, cividis, turbo
    alpha : float
        Alpha value (transparency).
    extended : bool
        If True, the lanthanoids and actinoids will be shown.
    periods_remove : List[int]
        Period numbers to be removed from the plot.
    groups_remove : List[int]
        Group numbers to be removed from the plot.
    log_scale : bool
        If True, the colorbar will be logarithmic.
    cbar_height : int
        Height of the colorbar.
    cbar_standoff : int
        Distance between the colorbar and the plot.
    cbar_fontsize : int
        Fontsize of the colorbar label.
    blank_color : str
        Hexadecimal color of the elements without data.
    under_value : float
        Values <= under_value will be colored with under_color.
    under_color : str
        Hexadecimal color to be used for the lower bound color.
    over_value : float
        Values >= over_value will be colored with over_color.
    under_color : str
        Hexadecial color to be used for the upper bound color.
    special_elements: List[str]
        List of elements to be colored with special_color.
    special_color: str
        Hexadecimal color to be used for the special elements.
    """

    values = [data[e] for e in data]

    if cmap == "viridis":
        cmap = mpl.cm.viridis
        bokeh_palette = "Viridis256"

    period_label = ["1", "2", "3", "4", "5", "6", "7"]
    group_range = [str(x) for x in range(1, 19)]

    period_label.append("blank")
    period_label.append("La")
    period_label.append("Ac")

    count = 0
    for i in range(56, 70):
        elements.period[i] = "La"
        elements.group[i] = str(count + 4)
        count += 1

    count = 0
    for i in range(88, 102):
        elements.period[i] = "Ac"
        elements.group[i] = str(count + 4)
        count += 1

    if log_scale:
        for datum in data:
            if datum < 0:
                raise ValueError(
                    f"Entry for element {datum} is negative but log-scale is selected"
                )
        color_mapper = LogColorMapper(
            palette=bokeh_palette, low=min(values), high=max(values)
        )
        norm = mpl.colors.LogNorm(vmin=min(values), vmax=max(values))
    else:
        color_mapper = LinearColorMapper(
            palette=bokeh_palette, low=min(values), high=max(values)
        )
        norm = mpl.colors.Normalize(vmin=min(values), vmax=max(values))
    color_scale = mpl.cm.ScalarMappable(norm=norm, cmap=cmap).to_rgba(
        values, alpha=None
    )

    # Set blank color
    color_list = [blank_color] * len(elements)
    # Compare elements in dataset with elements in periodic table
    for i, data_element in enumerate(data):
        element_entry = elements.symbol[
            elements.symbol.str.lower() == data_element.lower()
        ]
        if not element_entry.empty:
            element_index = element_entry.index[0]

        if under_value is not None and data[i] <= under_value:
            color_list[element_index] = under_color
        elif over_value is not None and data[i] >= over_value:
            color_list[element_index] = over_color
        else:
            color_list[element_index] = mpl.colors.to_hex(color_scale[i])

    if special_elements:
        for k, v in elements["symbol"].iteritems():
            if v in special_elements:
                color_list[k] = special_color

    # Define figure properties for visualizing data
    source = ColumnDataSource(
        data=dict(
            group=[str(x) for x in elements["group"]],
            period=[str(y) for y in elements["period"]],
            sym=elements["symbol"],
            atomic_number=elements["atomic number"],
            type_color=color_list,
        )
    )

    # Plot the periodic table
    p = figure(
        x_range=group_range, y_range=list(reversed(period_label)), tools="save"
    )
    p.plot_width = width
    p.outline_line_color = None
    p.background_fill_color = None
    p.border_fill_color = None
    p.toolbar_location = "above"
    p.rect(
        "group",
        "period",
        0.9,
        0.9,
        source=source,
        alpha=alpha,
        color="type_color",
    )
    p.axis.visible = False
    text_props = {
        "source": source,
        "angle": 0,
        "color": "black",
        "text_align": "left",
        "text_baseline": "middle",
    }
    x = dodge("group", -0.4, range=p.x_range)
    y = dodge("period", 0.3, range=p.y_range)
    p.text(
        x=x,
        y="period",
        text="sym",
        text_font_style="bold",
        text_font_size="16pt",
        **text_props,
    )
    p.text(x=x, y=y, text="atomic_number", text_font_size="11pt", **text_props)

    color_bar = ColorBar(
        color_mapper=color_mapper,
        ticker=BasicTicker(desired_num_ticks=10),
        border_line_color=None,
        label_standoff=cbar_standoff,
        location=(0, 0),
        orientation="vertical",
        scale_alpha=alpha,
        major_label_text_font_size=f"{cbar_fontsize}pt",
    )

    if cbar_height is not None:
        color_bar.height = cbar_height

    p.add_layout(color_bar, "right")
    p.grid.grid_line_color = None

    if output_filename:
        export_png(p, filename=output_filename)
    else:
        show_(p)
